"""Test the registry implementation and helpers."""
from dataclasses import dataclass
from typing import Optional

import pytest
from hopscotch import Registry
from hopscotch.fixtures.dataklasses import AnotherGreeting
from hopscotch.fixtures.dataklasses import Customer
from hopscotch.fixtures.dataklasses import FrenchCustomer
from hopscotch.fixtures.dataklasses import Greeter
from hopscotch.fixtures.dataklasses import GreeterCustomer
from hopscotch.fixtures.dataklasses import GreeterFirstName
from hopscotch.fixtures.dataklasses import GreeterFrenchCustomer
from hopscotch.fixtures.dataklasses import Greeting
from hopscotch.operators import context
from hopscotch.registry import IsNoneType
from hopscotch.registry import Registration


class DummyScan:
    """Mock out the venusian scanner."""

    def __init__(self):  # type: ignore
        """Fake called_with for use later."""
        self.called_with: Optional[object] = None

    def __call__(self, pkg: Optional[object]) -> None:
        """Set the called_with to simulate the caller."""
        self.called_with = pkg


def test_get_match_singleton_context_none() -> None:
    """Get best match when context is ``None``."""
    greeting = Greeting(salutation="no context")
    customer_greeting = Greeting(salutation="customer")

    # GIVEN A registry with two registrations
    r = Registry()  # No context
    r.register(greeting)
    r.register(customer_greeting, context=Customer)

    # WHEN Get the best match
    match = r.get_best_match(Greeting, allow_singletons=True)

    # THEN Match should be context=None registration
    if match:
        assert match.implementation == greeting


def test_nested_get_match_singleton_context_none() -> None:
    """Get best match from parent registry."""
    greeting = Greeting(salutation="no context")
    customer_greeting = Greeting(salutation="customer")

    # GIVEN A parent registry with two registrations, then a child.
    parent_registry = Registry()  # No context
    parent_registry.register(greeting)
    parent_registry.register(customer_greeting, context=Customer)
    child_registry = Registry(parent=parent_registry)

    # WHEN Get the best match
    match = child_registry.get_best_match(Greeting, allow_singletons=True)

    # THEN Match should be context=None registration
    if match:
        assert match.implementation == greeting


def test_singleton_registry_context_none() -> None:
    """When registry context is none, get correct singletons."""
    greeting = Greeting(salutation="no context")
    customer_greeting = Greeting(salutation="customer")

    # Singleton with context=None
    r = Registry(context=None)
    r.register(greeting)
    assert r.get(Greeting).salutation == "no context"
    # Singleton with context=Customer
    r = Registry(context=None)
    r.register(customer_greeting, context=Customer)
    with pytest.raises(LookupError):
        r.get(Greeting)


def test_singleton_registry_context_customer() -> None:
    """When registry context is ``Customer``, get correct singletons."""

    @dataclass()
    class NonCustomer:
        salutation: str = "Not a Customer"

    customer = Customer(first_name="customer")
    greeting = Greeting(salutation="no context")
    customer_greeting = Greeting(salutation="customer")
    french_greeting = Greeting(salutation="french customer")
    non_customer_greeting = Greeting(salutation="non customer")

    # ### A bunch of single-registration cases, no precedence involved
    # singleton.context=None
    r = Registry(context=customer)
    r.register(greeting)
    assert r.get(Greeting).salutation == "no context"

    # singleton.context=Customer means match
    r = Registry(context=customer)
    r.register(customer_greeting, context=Customer)
    assert r.get(Greeting).salutation == "customer"

    # singleton.context=FrenchCustomer means *no* match, too specific
    r = Registry(context=customer)
    r.register(french_greeting, context=FrenchCustomer)
    with pytest.raises(LookupError):
        r.get(Greeting)

    # singleton.context=NonCustomer means *no* match, not same type.
    r = Registry(context=customer)
    r.register(non_customer_greeting, context=NonCustomer)
    with pytest.raises(LookupError):
        r.get(Greeting)


def test_singleton_registry_context_french_customer() -> None:
    """When registry context is ``FrenchCustomer``, get correct singletons."""

    @dataclass()
    class NonCustomer:
        salutation: str = "Not a Customer"

    french_customer = FrenchCustomer(first_name="french customer")
    greeting = Greeting(salutation="no context")
    customer_greeting = Greeting(salutation="customer")
    french_greeting = Greeting(salutation="french customer")
    non_customer_greeting = Greeting(salutation="non customer")

    # ### A bunch of single-registration cases, no precedence involved
    # singleton.context=None
    r = Registry(context=french_customer)
    r.register(greeting)
    assert r.get(Greeting).salutation == "no context"

    # singleton.context=Customer means match
    r = Registry(context=french_customer)
    r.register(customer_greeting, context=Customer)
    assert r.get(Greeting).salutation == "customer"

    # singleton.context=FrenchCustomer matches
    r = Registry(context=french_customer)
    r.register(french_greeting, context=FrenchCustomer)
    assert r.get(Greeting).salutation == "french customer"

    # singleton.context=NonCustomer means *no* match, not same type.
    r = Registry(context=french_customer)
    r.register(non_customer_greeting, context=NonCustomer)
    with pytest.raises(LookupError):
        r.get(Greeting)


def test_field_operator_get_attr() -> None:
    """Use a field operator version of ``Get`` for the attribute."""
    r = Registry()
    r.register(GreeterFirstName)
    r.register(Customer(first_name="Field"))
    assert r.get(GreeterFirstName).customer_name == "Field"


def test_field_operator_context() -> None:
    """Use a field operator version of ``Context`` for the field."""
    r = Registry(context=FrenchCustomer(first_name="Marie"))
    r.register(GreeterFrenchCustomer)
    assert r.get(GreeterFrenchCustomer).customer.first_name == "Marie"


def test_singleton_registry_context_french_multiple() -> None:
    """Registry context is ``FrenchCustomer``, get singletons from multiple."""

    @dataclass()
    class NonCustomer:
        salutation: str = "Not a Customer"

    french_customer = FrenchCustomer(first_name="french customer")
    greeting = Greeting(salutation="no context")
    customer_greeting = Greeting(salutation="customer")
    french_greeting = Greeting(salutation="french customer")
    non_customer_greeting = Greeting(salutation="non customer")

    # ### A bunch of multiple-registration cases, no precedence involved
    # No context on registrations, choose last one.
    r = Registry(context=french_customer)
    r.register(greeting)
    r.register(Greeting(salutation="second"))
    assert r.get(Greeting).salutation == "second"

    # Matching the exact class is highest precedence
    r = Registry(context=french_customer)
    r.register(greeting)
    r.register(french_greeting, context=FrenchCustomer)
    r.register(customer_greeting, context=Customer)
    assert r.get(Greeting).salutation == "french customer"

    # Matching the exact class is highest precedence, registration order
    # doesn't matter
    r = Registry(context=french_customer)
    r.register(french_greeting, context=FrenchCustomer)
    r.register(greeting)
    r.register(customer_greeting, context=Customer)
    assert r.get(Greeting).salutation == "french customer"

    # Matching a subclass is highest precedence
    r = Registry(context=french_customer)
    r.register(greeting)
    r.register(customer_greeting, context=Customer)
    assert r.get(Greeting).salutation == "customer"

    # Matching the class is higher precedence, registration order
    # doesn't matter.
    r = Registry(context=french_customer)
    r.register(customer_greeting, context=Customer)
    r.register(greeting)
    assert r.get(Greeting).salutation == "customer"

    # Adding in a type that doesn't subclass, doesn't matter.
    r = Registry(context=french_customer)
    r.register(customer_greeting, context=Customer)
    r.register(non_customer_greeting, context=NonCustomer)
    r.register(greeting)
    assert r.get(Greeting).salutation == "customer"


def test_get_singleton() -> None:
    """Return a singleton that has no kind it is registered against."""
    registry = Registry()
    greeting = Greeting()
    registry.register(greeting)
    result = registry.get(Greeting)
    assert greeting is result


def test_get_singleton_kind() -> None:
    """Return a singleton that is registered against a kind."""
    registry = Registry()
    greeting = Greeting()
    registry.register(greeting, kind=Greeting)
    result = registry.get(Greeting)
    assert greeting.salutation == result.salutation


def test_get_singleton_kind_subclass() -> None:
    """Return a singleton that subclasses a kind."""
    registry = Registry()
    greeting = AnotherGreeting()
    registry.register(greeting, kind=Greeting)
    result = registry.get(Greeting)
    assert greeting is result


def test_get_kinds_found_class() -> None:
    """Construct an instance from a matching class."""
    registry = Registry()
    registry.register(AnotherGreeting)
    result = registry.get(Greeting)
    assert result.salutation == "Another Hello"


def test_get_kinds_match_in_parent() -> None:
    """No local match but is found in parent."""
    parent_registry = Registry()
    greeting = AnotherGreeting()
    parent_registry.register(AnotherGreeting)

    # Make a child registry which has nothing registered
    child_registry = Registry(parent=parent_registry)
    result = child_registry.get(Greeting)
    assert greeting == result


def test_nested_contexts() -> None:
    """Child registries get assigned the nearest context."""
    # We don't want to recurse up parents, looking for the
    # nearest non-None context.
    context = Customer(first_name="Grandparent")
    great_grandparent_registry = Registry()
    grandparent_registry = Registry(parent=great_grandparent_registry, context=context)
    parent_registry = Registry(parent=grandparent_registry)
    registry = Registry(parent=parent_registry)
    assert registry.context == context


def test_register_kind_with_class() -> None:
    """Register a singleton with the longer format."""
    registry = Registry()
    greeting = Greeting()
    registry.register(greeting)
    greetings = registry.registrations[Greeting]
    registration = greetings["singletons"][IsNoneType][0]
    assert registration.is_singleton
    assert registration.implementation == greeting
    assert registration.kind is None


def test_register_kind_without_class() -> None:
    """Register a singleton with the shorter format."""
    registry = Registry()
    greeting = AnotherGreeting()
    registry.register(greeting)
    gi = registry.registrations[AnotherGreeting]
    first = gi["singletons"][IsNoneType][0]
    assert first.implementation == greeting
    assert first.is_singleton


def test_register_class() -> None:
    """Register a class then ensure it is present."""
    registry = Registry()
    registry.register(AnotherGreeting)
    registrations = registry.registrations[Greeting]["classes"][IsNoneType]
    registration = registrations[0]
    assert AnotherGreeting is registration.implementation
    gs = registry.registrations[Greeting]
    first = gs["classes"][IsNoneType][0]
    assert first.implementation is AnotherGreeting
    assert first.kind is None
    assert not first.is_singleton


def test_register_kind_with_multiple_class() -> None:
    """Register a class as kind with implementations, get right one."""
    registry = Registry()
    registry.register(Greeting)
    registry.register(AnotherGreeting)
    result = registry.get(Greeting)
    assert result.__class__ is AnotherGreeting


def test_register_class_with_context() -> None:
    """Register a class for a context then ensure it is present."""
    registry = Registry()
    registry.register(AnotherGreeting, context=FrenchCustomer)
    classes = registry.registrations[Greeting]["classes"]
    registrations = classes[FrenchCustomer]
    registration = registrations[0]
    assert AnotherGreeting == registration.implementation
    gs = registry.registrations[Greeting]
    assert gs["classes"][IsNoneType] == []
    first = gs["classes"][FrenchCustomer][0]
    assert first.implementation is AnotherGreeting
    assert first.kind is None
    assert not first.is_singleton


def test_get_last_singleton_registration() -> None:
    """Register multiple singletons, last one wins."""
    g1 = Greeting(salutation="G1")
    g2 = Greeting(salutation="G2")
    registry = Registry()
    registry.register(g1)
    registry.register(g2)
    result = registry.get(Greeting)
    assert "G2" == result.salutation


def test_get_last_class_registration() -> None:
    """Register multiple classes, last one wins."""

    @dataclass()
    class GreetingImplementer2(Greeting):
        salutation: str = "G2"

    registry = Registry()
    registry.register(AnotherGreeting)
    registry.register(GreetingImplementer2)
    result = registry.get(Greeting)
    assert "G2" == result.salutation


def test_parent_registry() -> None:
    """Match a registry in the parent registry."""

    @dataclass()
    class GreetingImplementer2(Greeting):
        salutation: str = "G2"

    parent_registry = Registry()
    parent_registry.register(AnotherGreeting)
    child_registry = Registry(parent=parent_registry)
    child_registry.register(GreetingImplementer2, kind=Greeting)
    result = child_registry.get(Greeting)
    assert result.salutation == "G2"


def test_child_registry() -> None:
    """Match a registry in the child registry."""

    @dataclass()
    class GreetingImplementer2(Greeting):
        salutation: str = "G2"

    parent_registry = Registry()
    parent_registry.register(AnotherGreeting)
    parent_registry.register(GreetingImplementer2, kind=Greeting)
    child_registry = Registry(parent=parent_registry)
    result = child_registry.get(Greeting)
    assert "G2" == result.salutation


def test_callable_instance() -> None:
    """Match a registry in the child registry."""

    @dataclass()
    class ThisComponent:
        salutation: str = "This Greeting"

        def __call__(self) -> str:
            return f"Hello {self.salutation}"

    parent_registry = Registry()
    parent_registry.register(AnotherGreeting)
    parent_registry.register(ThisComponent, kind=Greeting)
    child_registry = Registry(parent=parent_registry)
    result = child_registry.get(Greeting)
    assert "This Greeting" == result.salutation


def test_nested_registry_match_parent() -> None:
    """Registration in parent uses dependency from parent."""

    @dataclass()
    class View:
        customer: Customer = context()

    context_customer = Customer(first_name="Parent")
    parent_registry = Registry(context=context_customer)
    parent_registry.register(View)
    child_registry = Registry(parent=parent_registry)
    result = child_registry.get(View)
    assert "Parent" == result.customer.first_name


def test_nested_registry_match_child() -> None:
    """Registration in parent uses dependency from child."""

    @dataclass()
    class View:
        customer: Customer = context()

    context_customer = Customer(first_name="Child")
    parent_registry = Registry()
    parent_registry.register(View)
    child_registry = Registry(parent=parent_registry, context=context_customer)
    result = child_registry.get(View)
    assert "Child" == result.customer.first_name


def test_dependency_not_in_registry() -> None:
    """Injection can call the symbol if it isn't registered."""
    from hopscotch.fixtures import plain_classes

    @dataclass()
    class GreeterPlainClassDependency:
        """A dataclass to engage a customer."""

        greeting: plain_classes.Greeting

    registry = Registry()
    registry.register(GreeterPlainClassDependency, kind=Greeter)
    greeter = registry.get(Greeter)
    assert greeter.greeting.salutation == "Hello"


def test_multiple_context_registrations() -> None:
    """Multiple registrations on different contexts."""
    customer = Customer(first_name="Fred")
    french_customer = FrenchCustomer(first_name="Marie")
    parent_registry = Registry()

    # Register for FrenchCustomer, Customer, and no context
    greeting = Greeting()
    parent_registry.register(greeting)
    parent_registry.register(Greeter)
    parent_registry.register(GreeterCustomer, kind=Greeter, context=Customer)
    parent_registry.register(
        GreeterFrenchCustomer, kind=Greeter, context=FrenchCustomer
    )

    # ## Now make per-request registries.
    # First, no context.
    request_registry = Registry(parent=parent_registry, context=None)
    greeter = request_registry.get(Greeter)
    assert isinstance(greeter, Greeter)

    # Next, context=Customer.
    request_registry = Registry(parent=parent_registry, context=customer)
    greeter = request_registry.get(Greeter)
    assert isinstance(greeter, GreeterCustomer)

    # Next, context=FrenchCustomer.
    request_registry = Registry(parent=parent_registry, context=french_customer)
    greeter = request_registry.get(Greeter)
    assert isinstance(greeter, GreeterFrenchCustomer)


def test_injector_registry_scan_caller() -> None:
    """Pretend to scan and see if called_with is set correctly."""
    registry = Registry()
    ds = DummyScan()  # type: ignore
    registry.scanner.scan = ds
    registry.scan()
    assert "tests" == ds.called_with.__name__  # type: ignore


def test_registration_with_context() -> None:
    """Ensure a registration can be created with a context."""
    registration = Registration(
        implementation=Greeting,
        context=Greeting,
    )
    assert registration.implementation is Greeting
    assert registration.kind is None
    assert registration.context is Greeting
    assert len(registration.field_infos) == 1
    assert not registration.is_singleton


def test_registration_with_kind() -> None:
    """Ensure a registration can be created with a kind."""
    registration = Registration(
        implementation=AnotherGreeting,
        kind=Greeting,
        context=Greeting,
    )
    assert registration.implementation is AnotherGreeting
    assert registration.kind is Greeting
    assert registration.context is Greeting
    assert len(registration.field_infos) == 1


def test_registration_singleton() -> None:
    """Ensure a singleton registration."""
    registration = Registration(
        implementation=Greeting,
        is_singleton=True,
    )
    assert registration.is_singleton


def test_registration_with_no_context() -> None:
    """Ensure a registration can be created without a context."""
    registration = Registration(implementation=Greeting)
    assert registration.implementation is Greeting
    assert registration.kind is None
    assert registration.context is None
    assert len(registration.field_infos) == 1


def test_context_registration_no_context() -> None:
    """A registration for a context in a registry with none."""
    registry = Registry()
    registry.register(Greeting, context=Customer)
    with pytest.raises(LookupError):
        registry.get(Greeting)


def test_registry_hopscotch_setup() -> None:
    """Module has ``hopscotch_setup`` called by ``registry.setup``."""
    from hopscotch.fixtures import hopscotch_setup
    from hopscotch.fixtures.hopscotch_setup import MyConfig

    registry = Registry()
    registry.setup(hopscotch_setup)
    my_config = registry.get(MyConfig)
    assert my_config.site_title == "My Configuration"



"""Example objects and kinds implemented as functions."""
from typing import Annotated
from typing import Optional

from hopscotch.fixtures.dataklasses import Customer
from hopscotch.operators import Get


def Greeting(salutation: str = "Hello") -> str:
    """A function to give a greeting."""
    return salutation


def GreetingDefaultNoHint(salutation="Hello") -> str:  # type: ignore
    """A function to with a parameter having no hint."""
    return salutation  # type: ignore


def GreetingNoDefault(salutation: str) -> str:
    """A function to give a greeting without a default."""
    return salutation


# Start Greeter
def Greeter(greeting: str) -> str:
    """A function to engage a customer."""
    return greeting


# Start GreeterOptional
def GreeterOptional(greeting: Optional[str]) -> Optional[str]:
    """A function to engage a customer with optional greeting."""
    return greeting


# Start GreeterAnnotated
def GreeterAnnotated(
    customer_name: Annotated[str, Get(Customer, attr="first_name")]
) -> str:
    """A function to engage a customer with an ``Annotated``."""
    return customer_name


def GreeterChildren(children: tuple[str]) -> tuple[str]:
    """A function that is passed a tree of VDOM nodes."""
    return children



"""Test the field discovery functions for various targets."""
import typing
from pathlib import Path

import pytest
from hopscotch import Registry
from hopscotch.field_infos import FieldInfo
from hopscotch.field_infos import get_dataclass_field_infos
from hopscotch.field_infos import get_field_origin
from hopscotch.field_infos import get_non_dataclass_field_infos
from hopscotch.field_infos import get_operator
from hopscotch.field_infos import get_stdlib_module_names
from hopscotch.fixtures import dataklasses
from hopscotch.fixtures import DummyOperator
from hopscotch.fixtures import functions
from hopscotch.fixtures import named_tuples
from hopscotch.fixtures import plain_classes
from hopscotch.fixtures.dataklasses import Customer
from hopscotch.fixtures.dataklasses import GreeterFirstName
from hopscotch.fixtures.dataklasses import Greeting
from hopscotch.fixtures.dataklasses import GreetingOperator
from hopscotch.fixtures.dataklasses import GreetingPath
from hopscotch.fixtures.dataklasses import GreetingTuple
from hopscotch.operators import Get
from hopscotch.operators import Operator


def test_get_stdlib_module_names() -> None:
    """Test the helper function which runs at startup."""
    module_names = get_stdlib_module_names()
    assert "contextlib" in module_names
    assert "builtins" in module_names


@pytest.mark.parametrize(
    "target, expected",
    [
        (dataklasses.Greeter, dataklasses.Greeting),
        (functions.Greeter, str),
        (named_tuples.Greeter, named_tuples.Greeting),
        (plain_classes.Greeter, plain_classes.Greeting),
    ],
)
def test_get_field_origin_no_generic(target: object, expected: object) -> None:
    """Find correct type with *no* generic involved."""
    th = typing.get_type_hints(target, globalns=None, localns=None)
    param1_annotation = th["greeting"]
    actual = get_field_origin(param1_annotation)
    assert expected is actual


@pytest.mark.parametrize(
    "target, expected",
    [
        (dataklasses.GreeterOptional, dataklasses.Greeting),
        (functions.GreeterOptional, str),
        (named_tuples.GreeterOptional, named_tuples.Greeting),
        (plain_classes.GreeterOptional, plain_classes.Greeting),
    ],
)
def test_get_field_origin_generic(target: object, expected: object) -> None:
    """Find correct type with a generic involved."""
    th = typing.get_type_hints(target, globalns=None, localns=None)
    param1_annotation = th["greeting"]
    result = get_field_origin(param1_annotation)
    assert expected is result


@pytest.mark.parametrize(
    "target, expected",
    [
        (dataklasses.GreeterAnnotated, dataklasses.Greeting),
        (named_tuples.GreeterAnnotated, named_tuples.Greeting),
        (plain_classes.GreeterAnnotated, plain_classes.Greeting),
    ],
)
def test_get_field_origin_annotated(target: object, expected: object) -> None:
    """Find correct type with an ``Annotated`` involved."""
    th = typing.get_type_hints(target, globalns=None, localns=None)
    param1_annotation = th["greeting"]
    result = get_field_origin(param1_annotation)
    assert expected is result


@pytest.mark.parametrize(
    "target, expected",
    [
        (functions.GreeterAnnotated, str),
    ],
)
def test_get_field_origin_annotated_function(target: object, expected: object) -> None:
    """Find correct type with an ``Annotated`` involved."""
    th = typing.get_type_hints(target, globalns=None, localns=None)
    param1_annotation = th["customer_name"]
    result = get_field_origin(param1_annotation)
    assert expected is result


@pytest.mark.parametrize(
    "target",
    [
        dataklasses.Greeting,
        functions.Greeting,
        named_tuples.Greeting,
        plain_classes.Greeting,
    ],
)
def test_get_field_origin_primitive_type(target: type) -> None:
    """Find correct type with a primitive type like ``str``."""
    th = typing.get_type_hints(target, globalns=None, localns=None)
    param1_annotation = th["salutation"]
    result = get_field_origin(param1_annotation)
    assert str is result


def test_parameter_no_hint() -> None:
    """A function parameter with no type hint should be None."""
    target = functions.GreetingDefaultNoHint
    field_infos = get_non_dataclass_field_infos(target)
    field_type = field_infos[0].field_type
    assert None is field_type


def test_dummy_get_operator() -> None:
    """See if we can get the annotation to get the operator."""
    aa = getattr(typing, "_AnnotatedAlias")
    field_type = aa(str, (DummyOperator("some_arg"),))
    field_type, operator = get_operator(field_type)
    assert str is field_type
    registry = Registry()
    if operator is not None:
        assert "some_arg" == operator(registry)


def test_get_operator_no_annotated() -> None:
    """A field with no ``Annotated`` shouldn't fail on get_operator.

    This simulates a field like ``title: str`` with no ``Annotated`` nor
    operator.
    """
    field_type, operator = get_operator(str)
    assert str is field_type
    assert None is operator


@pytest.mark.parametrize(
    "target, extractor",
    [
        (dataklasses.GreetingNoDefault, get_dataclass_field_infos),
        (functions.GreetingNoDefault, get_non_dataclass_field_infos),
        (named_tuples.GreetingNoDefault, get_non_dataclass_field_infos),
        (plain_classes.GreetingNoDefault, get_non_dataclass_field_infos),
    ],
)
def test_target_field_info_str(
    target: type, extractor: typing.Callable[..., list[FieldInfo]]
) -> None:
    """Variations of field_info extraction."""
    field_infos = extractor(target)
    assert field_infos[0].field_name == "salutation"
    assert field_infos[0].field_type == str
    assert field_infos[0].default_value is None
    assert field_infos[0].default_factory is None
    assert field_infos[0].has_annotated is False


@pytest.mark.parametrize(
    "target, extractor",
    [
        (dataklasses.GreeterChildren, get_dataclass_field_infos),
        (functions.GreeterChildren, get_non_dataclass_field_infos),
        (named_tuples.GreeterChildren, get_non_dataclass_field_infos),
        (plain_classes.GreeterChildren, get_non_dataclass_field_infos),
    ],
)
def test_field_info_children(
    target: type, extractor: typing.Callable[..., list[FieldInfo]]
) -> None:
    """Look for the magic-named ``children`` argument."""
    field_infos = extractor(target)
    assert field_infos[0].field_name == "children"
    assert field_infos[0].field_type is None
    assert field_infos[0].default_value is None


def test_field_info_default_value() -> None:
    """Confirm a default value."""
    field_infos = get_dataclass_field_infos(dataklasses.GreetingFieldDefault)
    assert field_infos[0].default_value == "Default Argument"


def test_field_info_default_factory() -> None:
    """Confirm a default factory."""
    field_infos = get_dataclass_field_infos(dataklasses.GreetingFieldDefaultFactory)
    assert field_infos[0].default_factory is list


def test_dataclass_field_info_init_false() -> None:
    """Look for the field constructor and init false."""
    field_infos = get_dataclass_field_infos(dataklasses.GreetingInitFalse)
    assert field_infos[0].field_name == "salutation"
    assert field_infos[0].field_type == str
    assert field_infos[0].default_value is None
    assert field_infos[0].has_annotated is False
    assert field_infos[0].is_builtin is True


def test_field_info_more_generic() -> None:
    """Look up a field that is a generic using ``tuple``."""
    field_infos = get_dataclass_field_infos(GreetingTuple)
    assert field_infos[0].field_name == "salutation"
    assert "GenericAlias" in str(type(field_infos[0].field_type))
    assert field_infos[0].default_value is None
    assert field_infos[0].is_builtin is True


def test_field_info_builtin() -> None:
    """Look up a field that is a a builtin ``Path``."""
    field_infos = get_dataclass_field_infos(GreetingPath)
    assert field_infos[0].field_name == "location"
    assert field_infos[0].field_type is Path
    assert field_infos[0].default_value is None
    assert field_infos[0].is_builtin is True


def test_dataclass_field_info_annotation() -> None:
    """Resolve an ``Annotated`` that uses an operator."""
    field_infos = get_dataclass_field_infos(GreetingOperator)
    assert field_infos[0].field_name == "greeter"
    assert field_infos[0].field_type is Greeting
    assert field_infos[0].default_value is None
    assert field_infos[0].is_builtin is False
    assert isinstance(field_infos[0].operator, Get)
    assert field_infos[0].has_annotated is True
    operator: Operator = field_infos[0].operator
    assert getattr(operator, "lookup_key") is Greeting


def test_dataclass_field_operator() -> None:
    """Field operators should result in same field info as annotated."""
    field_infos = get_dataclass_field_infos(GreeterFirstName)
    customer_name = field_infos[0]
    operator = customer_name.operator
    assert isinstance(operator, Get)
    assert operator.lookup_key == Customer
    assert operator.attr == "first_name"



"""Test usage of injectables via decorators instead of imperative.

More cumbersome (due to scanner) to copy around so placed into a
single test.
"""
from dataclasses import dataclass
from typing import cast

import pytest
from hopscotch import injectable
from hopscotch import Registration
from hopscotch import Registry
from hopscotch.fixtures import dataklasses
from hopscotch.fixtures.dataklasses import Customer
from hopscotch.fixtures.dataklasses import FrenchCustomer
from hopscotch.fixtures.dataklasses import Greeter
from hopscotch.fixtures.dataklasses import GreeterCustomer
from hopscotch.fixtures.dataklasses import GreeterFrenchCustomer


class View:
    """A marker for an example View custom injectable."""

    title: str


# noinspection PyPep8Naming
class view(injectable):  # noqa: N801
    """Custom decorator for custom injectable."""

    kind = View


class Config:
    """A marker for an example singleton custom injectable."""

    title: str


# noinspection PyPep8Naming
class config(injectable):  # noqa: N801
    """Custom decorator for custom singleton injectable."""

    kind = Config
    is_singleton = True


@view()
@dataclass
class MyView(View):
    """An example view."""

    title: str = "My View"


@config()
@dataclass
class MyConfig(View):
    """An example configuration."""

    title: str = "My Config"


def test_injectable_no_context() -> None:
    """The decorator and its registration do not care about context."""
    registry = Registry()
    registry.scan(dataklasses)
    greeter = registry.get(Greeter)
    assert greeter.greeting.salutation == "Hello"


def test_injectable_context() -> None:
    """The decorator has no context but the registration wants it."""
    customer = Customer(first_name="Fred")
    registry = Registry(context=customer)
    registry.scan(dataklasses)
    greeter = registry.get(GreeterCustomer)
    assert greeter.customer.first_name == "Fred"


def test_injectable_explicit_context() -> None:
    """A registration for a specific context."""
    french_customer = FrenchCustomer(first_name="Marie")

    # First, a registry that has a FrenchCustomer context.
    registry = Registry(context=french_customer)
    registry.scan(dataklasses)
    greeter = registry.get(GreeterFrenchCustomer)
    assert greeter.customer.first_name == "Marie"

    # This registry that has no FrenchCustomer context.
    registry2 = Registry()
    registry2.scan(dataklasses)
    with pytest.raises(LookupError):
        registry2.get(GreeterFrenchCustomer)


def test_custom_decorator() -> None:
    """Test the custom view decorator."""
    registry = Registry()
    registry.scan()
    result = cast(MyView, registry.get(View))
    assert "My View" == result.title


def test_custom_singleton_decorator() -> None:
    """Test the custom singleton decorator."""
    registry = Registry()
    registry.scan()
    config_singletons = registry.registrations[Config]["singletons"]
    items = cast(list, list(config_singletons.values())[0])
    first: Registration = items[0]
    impl = cast(MyConfig, first.implementation)
    assert "My Config" == impl.title
    result = cast(MyConfig, registry.get(Config))
    assert "My Config" == result.title



"""Sphinx configuration."""
from datetime import datetime

html_title = "Hopscotch"
project = "Hopscotch"
author = "Paul Everitt"
copyright = f"{datetime.now().year}, {author}"
extensions = [
    "myst_parser",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
]
autodoc_typehints = "description"
html_theme = "furo"
myst_enable_extensions = ["colon_fence"]



"""Setup Sybil and other central testing."""
from os import chdir
from os import getcwd
from shutil import rmtree
from tempfile import mkdtemp

import pytest
from sybil import Sybil
from sybil.parsers.codeblock import PythonCodeBlockParser
from sybil.parsers.doctest import DocTestParser


@pytest.fixture(scope="module")
def tempdir():
    """Wire up Sybil temp dir."""
    # there are better ways to do temp directories, but it's a simple example:
    path = mkdtemp()
    cwd = getcwd()
    try:
        chdir(path)
        yield path
    finally:
        chdir(cwd)
        rmtree(path)


pytest_collect_file = Sybil(
    parsers=[
        DocTestParser(),
        PythonCodeBlockParser(future_imports=["print_function"]),
    ],
    pattern="*.md",
    fixtures=["tempdir"],
).pytest()



"""Examples for the repo README.

The README has a "Quick Examples" section.
This is a tough place for examples:

- It's outside of Sphinx, to be served by GitHub
- But it's included into the top of the Sphinx docs

Because it isn't using Sphinx, we can't do includes.
Thus, we have to do cut-and-paste, so let's keep it to a minimum.
"""



"""Introspect target types and normalize to standard field info."""
from __future__ import annotations

import inspect
import os
import sysconfig
from dataclasses import Field
from dataclasses import fields
from dataclasses import is_dataclass
from dataclasses import MISSING
from inspect import Parameter
from inspect import signature
from typing import Any
from typing import Callable
from typing import get_args
from typing import get_origin
from typing import get_type_hints
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from hopscotch.operators import Operator

EMPTY = getattr(inspect, "_empty")


# At startup, make a set containing all the builtin module names
def get_stdlib_module_names() -> set[str]:
    """Called at startup to get a list of builtin modules."""
    stdlib_dir = sysconfig.get_path(name="stdlib")
    stdlib_modules = {f.replace(".py", "") for f in os.listdir(stdlib_dir)}
    # Add builtin "module" for things like tuple.
    stdlib_modules.add("builtins")
    return stdlib_modules


STDLIB_MODULE_NAMES = get_stdlib_module_names()


class FieldInfo(NamedTuple):
    """Extract needed info from dataclass fields, functions, etc."""

    field_name: str
    field_type: Optional[type]
    default_value: Optional[object] = None
    default_factory: Optional[Callable[[], object]] = None
    operator: Optional[Operator] = None
    has_annotated: bool = False
    is_builtin: bool = False


FieldInfos = list[FieldInfo]


def get_field_origin(field_type: Type[Any]) -> Any:
    """Helper to extract generic origin e.g. Optional[Resource]."""
    origin = get_origin(field_type)
    args = get_args(field_type)
    if origin is Union and args[-1] is type(None):  # noqa: E721
        return args[0]
    else:
        return field_type


def get_operator(field_type: Type[Any]) -> Tuple[type, Optional[Operator]]:
    """If using Annotation, get the operator."""
    operator = None
    if hasattr(field_type, "__metadata__"):
        field_type, operator = get_args(field_type)

    return field_type, operator


def dataclass_field_info_factory(field_type: type, field: Field[Any]) -> FieldInfo:
    """Return field metadata for a dataclass."""
    if field.name == "children":
        # Special case: a parameter named 'children'
        field_type = None
    else:
        # Is this a generic, such as Optional[KindContainer]?
        field_type = get_field_origin(field_type)

    # Using Annotation[] ??
    has_annotated = hasattr(field_type, "__metadata__")
    if has_annotated:
        field_type, operator = get_operator(field_type)
    else:
        # Field operators stash their info in field.metadata
        if "injected" in field.metadata:
            injected = field.metadata["injected"]
            operator = injected["operator"]
        else:
            operator = None

    # Default values
    default_value = None if field.default is MISSING else field.default
    default_factory = (
        None if field.default_factory is MISSING else field.default_factory
    )

    is_builtin = field_type.__module__ in STDLIB_MODULE_NAMES if field_type else False

    return FieldInfo(
        field_name=field.name,
        field_type=field_type,
        default_value=default_value,
        default_factory=default_factory,
        operator=operator,
        has_annotated=has_annotated,
        is_builtin=is_builtin,
    )


def function_field_info_factory(field_type: type, parameter: Parameter) -> FieldInfo:
    """Extract field info from a plain function."""
    operator = None
    is_builtin = False
    has_annotated = False

    this_field_type: Optional[type]
    if parameter.name == "children":
        # Special case: a parameter named 'children'
        this_field_type = None
    elif field_type == EMPTY:
        # If no type hint was provided, ``inspect._empty`` will be assigned.
        # Change this to None.
        this_field_type = None
    else:
        # Is this a generic, such as Optional[KindContainer]?
        this_field_type = get_field_origin(field_type)

        # Using Annotation[] ??
        if this_field_type:
            has_annotated = hasattr(this_field_type, "__metadata__")
            this_field_type, operator = get_operator(this_field_type)
            is_builtin = this_field_type.__module__ in STDLIB_MODULE_NAMES

    # Default values
    if parameter.default == EMPTY:
        default_value = None
    else:
        default_value = parameter.default

    return FieldInfo(
        field_name=parameter.name,
        field_type=this_field_type,
        default_value=default_value,
        operator=operator,
        has_annotated=has_annotated,
        is_builtin=is_builtin,
    )


def get_dataclass_field_infos(target: Callable[..., Any]) -> list[FieldInfo]:
    """Entry point to all sniffing at dataclasses."""
    type_hints = get_type_hints(
        target, include_extras=True, globalns=None, localns=None
    )
    # noinspection PyDataclass
    fields_mapping = {f.name: f for f in fields(target)}
    field_infos = [
        dataclass_field_info_factory(
            type_hints[field_name],
            fields_mapping[field_name],
        )
        for field_name in type_hints
    ]

    return field_infos


def get_non_dataclass_field_infos(target: Callable[..., Any]) -> list[FieldInfo]:
    """Entry point to all sniffing at non-dataclasses."""
    type_hints = get_type_hints(
        target, include_extras=True, globalns=None, localns=None
    )
    sig = signature(target)
    parameters = sig.parameters.values()
    field_infos = [
        function_field_info_factory(
            type_hints.get(param.name, param.annotation),
            param,
        )
        for param in parameters
    ]

    return field_infos


def get_field_infos(target: Any) -> FieldInfos:
    """Return field info for all the fields on a target."""
    if is_dataclass(target):
        return get_dataclass_field_infos(target)
    else:
        return get_non_dataclass_field_infos(target)



"""Utility functions for scanning packages and modules."""
import sys
from types import ModuleType
from typing import Callable


def caller_module(level: int = 2) -> ModuleType | None:
    """Return the module of the caller of the current scope."""
    getframe = getattr(sys, "_getframe")
    module_globals = getframe(level).f_globals
    module_name = module_globals.get("__name__") or "__main__"
    if module_name == "__test__":
        # We are running from a doctest via Sybil so
        # don't try to sniff around
        return None
    module: ModuleType = sys.modules[module_name]
    return module


def caller_package(
    level: int = 2,
    cm: Callable[
        [
            int,
        ],
        ModuleType,
    ] = caller_module,
) -> ModuleType | None:
    """Return the package for the caller of a certain scope."""
    module = cm(level + 1)
    if module is None:
        # We are in a doctest via Sybl
        return None
    f = getattr(module, "__file__", "")
    if ("__init__.py" in f) or ("__init__$py" in f):  # empty at >>>
        # Module is a package
        return module
    # Go up one level to get package
    package_name = module.__name__.rsplit(".", 1)[0]
    return sys.modules[package_name]



"""Test suite for caller utility functions."""
from hopscotch.callers import caller_module
from hopscotch.callers import caller_package
from hopscotch.fixtures.init_caller_package import main_package


def test_caller_module() -> None:
    """Ensure the utility function returns the calling module."""
    result = caller_module()
    assert "_pytest.python" == result.__name__


def test_caller_package() -> None:
    """Ensure the utility function returns the calling package."""
    result = caller_package()
    assert "_pytest" == result.__name__


def test_caller_package_init() -> None:
    """Ensure the utility function returns the calling package init."""
    result = main_package()
    assert result.__name__ == "hopscotch.fixtures.init_caller_package"



"""Default operators for fields.

Operators allow a field to ask for some work to be done during injection.
"""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import Field
from dataclasses import field
from typing import Any
from typing import Optional
from typing import Protocol
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .registry import Registry


class Operator(Protocol):
    """Specify the structure of operator implementations."""

    def __call__(self, registry: Registry) -> str:
        """Operators are callables."""
        ...


def make_field_operator(operator_class: Any) -> Any:
    """Helper to convert an Operator into a dataclass field."""

    def _inner(*args, **kwargs) -> Field:  # type: ignore
        # Some of the kwargs are for generic field keyword
        # arguments, such as ``init=False``. But some should be
        # sent to the operator. Put them in two piles.
        operator_kwargs = {}
        for kwarg_key, kwarg_value in list(kwargs.items()):
            if kwarg_key not in Field.__slots__:
                operator_kwargs[kwarg_key] = kwarg_value
                del kwargs[kwarg_key]

        # We can now construct the operator
        operator = operator_class(*args, **operator_kwargs)
        if "metadata" not in kwargs:
            kwargs["metadata"] = {}

        # Use dataclass field metadata support to smuggle our injector
        # information through to the other side.
        kwargs["metadata"]["injected"] = {"operator": operator}
        return field(**kwargs)  # type: ignore

    return _inner


# Start Get
@dataclass(frozen=True)
class Get:
    """Lookup a kind and optionally pluck an attr."""

    lookup_key: Any
    attr: Optional[str] = None

    def __call__(
        self,
        registry: Registry,
    ) -> object:
        """Use registry to find lookup key and optionally pluck attr."""
        # Can't lookup a string, ever, so bail on this with an error.
        if isinstance(self.lookup_key, str):
            lk = self.lookup_key
            msg = f"Cannot use a string {lk!r} as container lookup value"
            raise ValueError(msg)

        result_value = registry.get(self.lookup_key)

        # Are we plucking an attr?
        if self.attr is not None:
            result_value = getattr(result_value, self.attr)

        return result_value


get = make_field_operator(Get)


@dataclass(frozen=True)
class Context:
    """Grab the current container context and optionally pluck an attr."""

    attr: Optional[str] = None

    def __call__(
        self,
        registry: Registry,
    ) -> object:
        """Use registry to grab the context and optionally pluck an attr."""
        value = registry.context
        if value is None:
            raise ValueError("No context on registry")

        # Are we plucking an attr?
        if self.attr is not None:
            value = getattr(value, self.attr)

        return value


context = make_field_operator(Context)



"""Example objects for tests, examples, and docs."""
from dataclasses import dataclass

from hopscotch.registry import Registry


@dataclass()
class DummyOperator:
    """Simulate an operator that looks something up."""

    arg: str

    def __call__(self, registry: Registry) -> str:
        """Return the stored argument.

        Args:
            registry: A real or fake hopscotch registry for this operator.

        Returns:
            The result of this DummyOperator.
        """
        return self.arg



"""Test the example."""


def test_readme_context() -> None:
    """We got a different greeter based on context."""
    from . import greeter1, greeter2

    assert greeter1.greeting == "Hello!"
    assert greeter2.greeting == "Bonjour!"



"""Package wants to manually setup the registry."""
from dataclasses import dataclass

from hopscotch import Registry


@dataclass
class MyConfig:
    """Example configuration class for this fake site."""

    site_title: str


def hopscotch_setup(registry: Registry) -> None:
    """Manually configure the registry for this package."""
    my_config = MyConfig(site_title="My Configuration")
    registry.register(my_config)



"""Hopscotch."""
from __future__ import annotations

from hopscotch.registry import inject_callable
from hopscotch.registry import injectable
from hopscotch.registry import Registration
from hopscotch.registry import Registry

__all__ = [
    "Registry",
    "injectable",
    "inject_callable",
    "Registration",
]



"""Register with a decorator."""
from dataclasses import dataclass

from hopscotch import injectable
from hopscotch import Registry


@injectable()
@dataclass
class Greeter:
    """A simple greeter."""

    greeting: str = "Hello!"


registry = Registry()
registry.scan()
# Later
greeter = registry.get(Greeter)
# greeter.greeting == "Hello!"



"""Test the injection process.

The ``inject_callable`` callable is used in both the registry and
components. Thus it needs to support use both with and without a
registry.
"""
from dataclasses import dataclass
from dataclasses import field

import pytest
from hopscotch import Registry
from hopscotch.fixtures import functions
from hopscotch.fixtures import named_tuples
from hopscotch.fixtures.dataklasses import AnotherGreeting
from hopscotch.fixtures.dataklasses import Customer
from hopscotch.fixtures.dataklasses import Greeter
from hopscotch.fixtures.dataklasses import GreeterCustomer
from hopscotch.fixtures.dataklasses import GreeterKind
from hopscotch.fixtures.dataklasses import GreeterRegistry
from hopscotch.fixtures.dataklasses import Greeting
from hopscotch.fixtures.dataklasses import GreetingFactory
from hopscotch.fixtures.dataklasses import GreetingFieldDefault
from hopscotch.fixtures.dataklasses import GreetingFieldDefaultFactory
from hopscotch.fixtures.dataklasses import GreetingNoDefault
from hopscotch.registry import inject_callable
from hopscotch.registry import Registration


def test_field_default() -> None:
    """The target is a field with a default."""
    registration = Registration(Greeting)
    result: Greeting = inject_callable(registration)
    assert result.salutation == "Hello"


def test_field_default_argument() -> None:
    """The target is a field using a default argument."""
    registration = Registration(GreetingFieldDefault)
    result: Greeting = inject_callable(registration)
    assert result.salutation == "Default Argument"


def test_field_default_factory() -> None:
    """The target is a field using a default factory."""
    registration = Registration(GreetingFieldDefaultFactory)
    result: Greeting = inject_callable(registration)
    assert result.salutation == []


def test_dependency_class() -> None:
    """The target has a field dependency to fetch from registry."""
    registry = Registry()
    registry.register(Greeting)

    registration = Registration(GreeterKind)
    result: GreeterKind = registry.inject(registration)
    assert "Hello" == result.greeting.salutation


def test_dependency_default_class() -> None:
    """The target has a field dependency NOT in registry."""

    @dataclass
    class SomeSalutation:
        title: str = "Some Salutation"

    @dataclass
    class SomeGreeting:
        some_salutation: SomeSalutation = field(default_factory=SomeSalutation)

    registry = Registry()

    registry.register(SomeGreeting)
    registry.register(SomeGreeting)
    registration = Registration(SomeGreeting)
    result: SomeGreeting = registry.inject(registration)
    assert "Some Salutation" == result.some_salutation.title


def test_dependency_namedtuple() -> None:
    """Inject a non-type that isn't registered.

    We use ``registry.get`` to get an implementation that is a type of
    the thing we are looking up. Things like ``NamedTuple`` and
    functions can't be subclasses.

    They can still be used by the injector, just by grabbing the symbol
    directly rather than going to find it.
    """
    registration = Registration(named_tuples.Greeter)
    result: Greeter = inject_callable(registration)
    assert "Hello" == result.greeting.salutation


def test_inject_function_no_type_hint() -> None:
    """A function parameter with no type hint can use default value."""
    registration = Registration(functions.GreetingDefaultNoHint)
    result: str = inject_callable(registration)
    assert "Hello" == result


def test_injection_no_registry() -> None:
    """Simulate usage of injection rules without needing a registry."""
    props = {"salutation": "No registry"}
    registration = Registration(Greeting)
    result: Greeting = inject_callable(registration, props=props)
    assert "No registry" == result.salutation


def test_dependency_no_default() -> None:
    """The target has str field with no default, fail with custom exception."""
    registration = Registration(GreetingNoDefault)
    with pytest.raises(ValueError) as exc:
        inject_callable(registration)

    expected = "Cannot inject 'str' on 'GreetingNoDefault.salutation'"
    assert exc.value.args[0] == expected


def test_dependency_default() -> None:
    """The target has an str field with a default."""
    registration = Registration(AnotherGreeting)
    result: AnotherGreeting = inject_callable(registration)
    assert "Another Hello" == result.salutation


def test_non_dependency() -> None:
    """The target has a dependent field to fetch from registry."""
    gs = Greeting(salutation="use singleton")
    registry = Registry()
    registry.register(gs)
    registration = Registration(Greeter)
    result: Greeter = inject_callable(registration, registry=registry)
    assert "use singleton" == result.greeting.salutation


def test_dependency_nested_registry() -> None:
    """Nested registry, can injector get singleton from right level?"""
    gs_child = Greeting(salutation="use child")
    gs_parent = Greeting(salutation="use parent")

    # Site registry
    parent_registry = Registry()
    parent_registry.register(gs_parent)

    # Per-request registry with a specific singleton
    child_registry = Registry(parent=parent_registry)
    child_registry.register(gs_child)

    # Get something registered with parent, dependency local
    registration = Registration(GreeterKind)
    result: GreeterKind = child_registry.inject(registration)
    assert "use child" == result.greeting.salutation


def test_pass_in_props_create_dependency() -> None:
    """Instead of injecting a field, get it from passed-in 'props'."""
    props = dict(salutation="use prop")  # noqa: C408
    registration = Registration(Greeting)
    result: Greeting = inject_callable(registration, props=props)
    assert result.salutation == "use prop"


def test_inject_registry() -> None:
    """Target wants the registry and will later get what it needs."""
    registry = Registry()
    registration = Registration(GreeterRegistry)
    result: GreeterRegistry = inject_callable(registration, registry=registry)
    assert registry == result.registry


def test_inject_context() -> None:
    """Target wants the registry context."""
    customer = Customer(first_name="Mary")
    registry = Registry(context=customer)
    registration = Registration(GreeterCustomer)
    registry.register(GreeterCustomer)
    result: GreeterCustomer = inject_callable(registration, registry=registry)
    assert customer == result.customer


def test_hopscotch_factory() -> None:
    """The dependency has its own factory as a class attribute."""
    r = Registration(GreetingFactory)
    registry = Registry()
    registry.register(GreetingFactory)
    result: GreetingFactory = inject_callable(r, registry=registry)
    assert result.salutation == "Hi From Factory"



"""Replace a built-in class."""
from dataclasses import dataclass

from hopscotch import injectable
from hopscotch import Registry


@dataclass
class Greeter:
    """A simple greeter."""

    greeting: str = "Hello!"


# Some site might want to change a built-in.
@injectable(kind=Greeter)
@dataclass
class CustomGreeter:
    """Provide a different ``Greeter`` in this site."""

    greeting: str = "Howdy!"


# Make app and register built-ins
registry = Registry()
registry.register(Greeter)
# Still during startup, but later during site registration
registry.scan()
# Later, when processing a request
greeter = registry.get(Greeter)
# greeter.greeting == "Howdy!"



"""Rich dependency injection with operators."""
from dataclasses import dataclass

from hopscotch import injectable
from hopscotch import Registry
from hopscotch.operators import get


@injectable()
@dataclass
class SiteConfig:
    """Hold configuration data for a pluggable app site."""

    punctuation: str = "!"


@injectable()
@dataclass
class Greeter:
    """A simple greeter."""

    punctuation: str = get(SiteConfig, attr="punctuation")
    greeting: str = "Hello"

    def greet(self) -> str:
        """Provide a greeting."""
        return f"{self.greeting}{self.punctuation}"


registry = Registry()
registry.scan()
# Later
greeter = registry.get(Greeter)
greeting = greeter.greet()
# greeting == "Hello!"



"""Type-oriented registry that start simple and finishes powerful."""
from __future__ import annotations

from collections import defaultdict
from dataclasses import dataclass
from dataclasses import field
from importlib import import_module
from inspect import getmro
from inspect import isclass
from types import ModuleType
from typing import Any
from typing import Callable
from typing import cast
from typing import Optional
from typing import Type
from typing import TypedDict
from typing import TypeVar
from typing import Union

from venusian import attach
from venusian import Scanner

from .callers import caller_package
from .field_infos import FieldInfo
from .field_infos import FieldInfos
from .field_infos import get_field_infos

PACKAGE = Optional[Union[ModuleType, str]]
Props = dict[str, Any]


class IsNoneType:
    """Mimic Python 3.10 ``NoneType`` as just a marker."""

    pass


@dataclass()
class Registration:
    """Collect registration and introspection info of a target."""

    slots = ("implementation", "kind", "context", "field_infos", "is_singleton")
    implementation: Union[Callable[..., object], object]
    kind: Optional[Callable[..., object]] = None
    context: Optional[Callable[..., object]] = None
    field_infos: FieldInfos = field(default_factory=list)
    is_singleton: bool = False

    def __post_init__(self) -> None:
        """Extract and assign the field infos if not singleton."""
        if not self.is_singleton:
            self.field_infos = get_field_infos(self.implementation)


T = TypeVar("T")


def inject_field_no_registry(
    field_info: FieldInfo,
    props: Optional[Props],
) -> Optional[object]:
    """Get a value for a field without a registry."""
    ft = field_info.field_type
    is_builtin = field_info.is_builtin
    if not (ft is None or is_builtin):
        # Avoid trying to inject str, meaning, only inject
        # user-defined classes
        # Treat this as a symbol that can be injected without
        # a lookup, such as a function, NamedTuple, etc.
        # TODO Would be great to avoid computing this each time
        registration = Registration(
            context=None,
            implementation=ft,
            is_singleton=False,  # TODO They might be in registry
        )
        return inject_callable(registration, props=props)

    return None


def inject_field_registry(
    field_info: FieldInfo,
    registry: Registry,
) -> Optional[Any]:
    """Get a value for a field using a registry."""
    ft = field_info.field_type
    is_builtin = field_info.is_builtin
    operator = field_info.operator
    if operator is not None:
        # This field uses Annotated[SomeType, SomeOperator]
        return operator(registry)
    elif ft is Registry:
        # Special rule: if you ask for the registry, you'll get it
        return registry
    elif not (ft is None or is_builtin):
        # Avoid trying to inject str, meaning, only inject
        # user-defined classes
        # TODO If ``ft`` is a function or NamedTuple, it kind of breaks
        #   the type-oriented contract for ``get``. But the following
        #   might still work.  Not sure the right solution.
        try:
            return registry.get(ft)
        except LookupError as exc:
            # See if this dependency is in parent registries
            if registry.parent:
                return registry.parent.get(ft)
            # No parent registry, just re-raise the exception
            raise exc

    return None


def inject_callable(
    registration: Registration,
    props: Optional[Props] = None,
    registry: Optional[Registry] = None,
) -> T:
    """Construct target with or without a registry."""
    target = registration.implementation
    kwargs = {}

    # If the target has a ``__hopscotch_factory__``, use that instead
    # of automated construction.
    factory = getattr(target, "__hopscotch_factory__", None)

    if factory is not None and registry is not None:
        result: T = factory(registry)
        return result

    for field_info in registration.field_infos:
        fn = field_info.field_name
        if props and fn in props:
            # Props have highest precedence
            field_value = props[fn]
        elif registry:
            try:
                field_value = inject_field_registry(field_info, registry)
            except LookupError:
                # During *injection* (not during ``registry.get``) we
                # allow injectable dependencies that aren't registered.
                # Maybe a function, dataclass, whatever. Just inject it.
                field_value = inject_field_no_registry(field_info, props)
        else:
            field_value = inject_field_no_registry(field_info, props)

        # If we didn't get a value...
        if field_value is None:
            if field_info.default_value is not None:
                # ...try to get from default
                field_value = field_info.default_value
            elif field_info.default_factory is not None:
                field_value = field_info.default_factory()
            else:
                # ...otherwise, we failed injection.
                ql = target.__qualname__  # type: ignore
                ft = field_info.field_type
                ft_name = "None" if ft is None else ft.__name__
                msg = f"Cannot inject {ft_name!r} on '{ql}.{fn}'"
                raise ValueError(msg)

        kwargs[fn] = field_value

    # Construct and return the class
    return target(**kwargs)  # type: ignore


class KindGroups(TypedDict):
    """Constrain the keys to just singleton and classes."""

    singletons: dict[Union[type, IsNoneType], list[Registration]]
    classes: dict[Union[type, IsNoneType], list[Registration]]


def make_singletons_classes() -> KindGroups:
    """Factory for defaultdict to initialize second level of tree."""
    kind_groups: KindGroups = {
        "singletons": defaultdict(list),
        "classes": defaultdict(list),
    }
    return kind_groups


Registrations = dict[type, KindGroups]


class Registry:
    """Type-oriented registry with special features."""

    context: Optional[Any]
    parent: Optional[Registry]
    scanner: Scanner
    registrations: Registrations

    def __init__(
        self,
        parent: Optional[Registry] = None,
        context: Optional[Any] = None,
    ) -> None:
        """Construct a registry that might have a context and be nested."""
        self.registrations = defaultdict(make_singletons_classes)
        self.parent: Optional[Registry] = parent
        if context is None and parent is not None:
            self.context = parent.context
        else:
            self.context = context
        self.scanner = Scanner(registry=self)

    def setup(
        self,
        pkg: PACKAGE = None,
    ) -> None:
        """Pass the registry to a package which has a setup function."""
        setup_function = getattr(pkg, "hopscotch_setup", None)
        if setup_function:
            setup_function(self)

    def scan(
        self,
        pkg: PACKAGE = None,
    ) -> None:
        """Look for decorators that need to be registered."""
        if pkg is None:
            # Get the caller module and import it
            pkg = caller_package()
        elif isinstance(pkg, str):
            # importlib.resource package specification
            pkg = import_module(pkg)
        if pkg is not None:
            self.scanner.scan(pkg)

    def inject(self, registration: Registration, props: Optional[Props] = None) -> T:
        """Use injection to construct and return an instance."""
        return inject_callable(registration, props=props, registry=self)

    def get_best_match(
        self,
        kind: Type[T],
        context_class: Optional[Any] = None,
        allow_singletons: bool = True,  # If props are passed in, we can't use singletons
    ) -> Optional[Registration]:
        """Find the best-match registration, if any.

        Using the registry is a two-step process: lookup an implementation,
        then if needed, construct and return. This is the first part.
        """
        tr = self.registrations[kind]
        if allow_singletons:
            registrations = tr["singletons"] | tr["classes"]
        else:
            registrations = tr["classes"]

        # We will put possible matches into three piles: high/medium/low
        # precedence. Each pile is an ordered list based on registration
        # time (later registrations override earlier.)
        precedences: dict[str, list[Registration]] = {
            "high": [],
            "medium": [],
            "low": [],
        }
        for this_context, these_registrations in registrations.items():
            if this_context is IsNoneType and context_class is None:
                # This is the most basic case, test it first to bail out quickly.
                precedences["low"] = these_registrations
            elif this_context is context_class:
                precedences["high"] = these_registrations
            elif this_context is IsNoneType:
                precedences["low"] = these_registrations
            if this_context is not IsNoneType:
                # TODO mypy didn't like issubclass(context_class, this_context)
                if context_class and issubclass(
                    context_class, cast(type, this_context)
                ):
                    precedences["medium"] = these_registrations
            # Otherwise, filter it out and do nothing

        # Get the first match and return (is_singleton=True) or construct
        # and return (is_singleton = False)
        matches = precedences["high"] + precedences["medium"] + precedences["low"]

        # If we found a match, return it
        if matches:
            return matches[0]
        else:
            if self.parent:
                # Otherwise, go to parent
                return self.parent.get_best_match(
                    kind,
                    context_class=context_class,
                    allow_singletons=allow_singletons,
                )
            return None

    def get(  # noqa: C901
        self,
        kind: Type[T],
        context: Optional[Any] = None,
        **kwargs: Props,
    ) -> T:
        """Find an appropriate kind class and construct an implementation.

        The passed-in keyword args act as "props" which have highest-precedence
        as arguments used in construction.
        """
        # Use the passed-in context class if provided, otherwise, the
        # the registry's context (if provided.)
        context_class: Optional[Any] = None
        if context:
            context_class = context.__class__
        elif self.context:
            context_class = self.context.__class__

        # Use precedence etc. to get the best matching implementation.
        best_match = self.get_best_match(
            kind,
            context_class=context_class,
            # If props are passed in, we can't use singletons. So
            # allow_singletons when bool(kwargs) is false.
            allow_singletons=not bool(kwargs),
        )
        if best_match:
            if best_match.is_singleton:
                # TODO Now that ``.get()`` makes promises about typing, it makes
                #    registering a singleton for a "kind" a good bit harder.
                #    We'll just shut this up for now.
                return best_match.implementation  # type: ignore
            else:
                # Need to construct it
                instance: T = self.inject(best_match, props=kwargs)
                return instance

        # If we get to here, we didn't find anything, raise an error
        msg = f"No kind {kind.__name__!r} in registry"
        raise LookupError(msg)

    def register(
        self,
        implementation: T,
        *,
        kind: Optional[Type[T]] = None,
        context: Optional[Any] = None,
    ) -> None:
        """Use a LIFO list for all the possible implementations.

        Note that the implementation must be a subclass of the kind.
        """
        is_singleton = not isclass(implementation)

        registration = Registration(
            implementation=implementation,
            context=context,
            kind=kind,
            is_singleton=is_singleton,
        )

        # Let's decide what key to use to register this as.
        if kind is None:
            # Let's try to infer it from the subclass.
            if isclass(implementation):
                base_classes = getmro(implementation)[:-1]
                if len(base_classes) > 1:
                    # This registration is a class with a single base class
                    st = base_classes[1]
                else:
                    # Mimic the part below to finish this branch
                    st = implementation
            else:
                st = type(implementation) if is_singleton else implementation
        else:
            st = kind

        # Put this in the correct place of the registrations tree,
        # creating tree nodes as needed.

        s_or_c = "singletons" if is_singleton else "classes"
        registrations = self.registrations[st][s_or_c]  # type: ignore
        this_context = IsNoneType if context is None else context
        registrations[this_context].insert(0, registration)


class injectable:  # noqa
    """``venusian`` decorator to register an injectable factory ."""

    kind: Optional[
        Type[T]
    ] = None  # Give subclasses a chance to give default, e.g. view
    is_singleton: bool = False  # Decorator registers singletons.

    def __init__(
        self,
        kind: Optional[Type[T]] = None,
        *,
        context: Optional[Optional[Any]] = None,
    ):
        """Construct decorator that can register later with registry."""
        if kind:
            self.kind = kind
        self.context = context

    def __call__(self, wrapped: T) -> T:
        """Execute the decorator during venusian scan phase."""

        def callback(scanner: Scanner, name: str, cls: type) -> None:
            """Perform the work of actually putting in registry."""
            # Does this custom decorator want to register singletons? If
            # so, make an instance to use instead of decorated class target.
            target = cls() if self.is_singleton else cls
            registry = getattr(scanner, "registry")
            registry.register(
                implementation=target,
                kind=self.kind,
                context=self.context,
            )

        attach(wrapped, callback)
        return wrapped



"""Example objects and kinds implemented as dataclasses."""
from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from pathlib import Path
from typing import Annotated
from typing import Optional

from hopscotch import injectable
from hopscotch import Registry
from hopscotch.operators import Context
from hopscotch.operators import context
from hopscotch.operators import Get
from hopscotch.operators import get


# Decorate Greeting
@injectable()
@dataclass()  # Start Greeting
class Greeting:
    """A dataclass to give a greeting."""

    salutation: str = "Hello"


# Start AnotherGreeting
@dataclass()
class AnotherGreeting(Greeting):
    """A replacement alternative for the default ``Greeting``."""

    salutation: str = "Another Hello"


@dataclass()
class GreetingNoDefault:
    """A dataclass to give a greeting with no default value."""

    salutation: str


@dataclass()
class GreetingInitFalse:
    """A dataclass with a field that inits to false."""

    salutation: str = field(init=False)


@dataclass()
class GreetingFieldDefault:
    """A dataclass with a field using a default argument."""

    salutation: str = field(default="Default Argument")


@dataclass()
class GreetingFieldDefaultFactory:
    """A dataclass with a field using a default factory."""

    salutation: list = field(default_factory=list)


@dataclass()
class GreetingOperator:
    """A dataclass to give a greeting via an operator."""

    greeter: Annotated[
        Greeting,
        Get(Greeting),
    ]


@dataclass()
class GreetingTuple:
    """A dataclass to give a sequence of greetings."""

    salutation: tuple[str, ...]


@dataclass()
class GreetingPath:
    """A dataclass to give a builtin Path."""

    location: Path


# Start GreetingFactory
@dataclass()
class GreetingFactory:
    """Use the ``__hopscotch_factory__`` protocol to control creation."""

    salutation: str

    @classmethod
    def __hopscotch_factory__(cls, registry: Registry) -> GreetingFactory:
        """Manually construct this instance, instead of injection."""
        return cls(salutation="Hi From Factory")


# Start Greeter
@injectable()
@dataclass()
class Greeter:
    """A dataclass to engage a customer."""

    greeting: Greeting


# Start GreeterKind
@dataclass()
class GreeterKind:
    """A dataclass ``Kind`` to engage a customer."""

    greeting: Greeting


# Start GreeterOptional
@dataclass()
class GreeterOptional:
    """A dataclass to engage a customer with optional greeting."""

    greeting: Optional[Greeting]  # no default


@dataclass()
class GreeterAnnotated:
    """A dataclass to engage a customer with an annotation."""

    greeting: Annotated[Greeting, "YOLO"]


@dataclass()
class GreeterChildren:
    """A dataclass that is passed a tree of VDOM nodes."""

    children: tuple[str]


@dataclass()
class GreeterRegistry:
    """A dataclass that depends on the registry."""

    registry: Registry


# Start Customer
@dataclass()
class Customer:
    """The person to greet, stored as the registry context."""

    first_name: str


@dataclass()
class FrenchCustomer(Customer):
    """A different kind of person to greet, stored as the registry context."""

    first_name: str


@injectable()
@dataclass()
class GreeterCustomer:
    """A dataclass that depends on the registry context."""

    customer: Annotated[Customer, Context()]


# Start GreeterFrenchCustomer
@injectable(context=FrenchCustomer)
@dataclass()
class GreeterFrenchCustomer:
    """A dataclass that depends on a different registry context."""

    customer: FrenchCustomer = context()


# Start GreeterGetAnother
@injectable()
@dataclass()
class GreeterGetAnother:
    """Use an operator to change the type hint of what's retrieved."""

    customer_name: AnotherGreeting = get(Greeting)


# Start GreeterFirstName
@dataclass()
class GreeterFirstName:
    """A dataclass that gets an attribute off a dependency."""

    customer_name: str = get(Customer, attr="first_name")



"""Example objects and kinds implemented as ``NamedTuple``.

Note that, since ``typing.NamedTuple`` doesn't really do inheritance,
we can't implement a ``Kind`` as a ``NamedTuple``.
"""
from typing import Annotated
from typing import NamedTuple
from typing import Optional


class Greeting(NamedTuple):
    """A ``NamedTuple`` to give a greeting."""

    salutation: str = "Hello"


class GreetingNoDefault(NamedTuple):
    """A ``NamedTuple`` to give a greeting without a default."""

    salutation: str


# Start Greeter
class Greeter(NamedTuple):
    """A ``NamedTuple`` to engage a customer."""

    greeting: Greeting


class GreeterOptional(NamedTuple):
    """A ``NamedTuple`` to engage a customer with optional greeting."""

    greeting: Optional[Greeting]


class GreeterAnnotated(NamedTuple):
    """A ``NamedTuple`` to engage a customer with an annotation."""

    greeting: Annotated[Greeting, "YOLO"]


class GreeterChildren(NamedTuple):
    """A ``NamedTuple`` that is passed a tree of VDOM nodes."""

    children: tuple[str]



"""Test the bundled operators."""
from dataclasses import dataclass
from typing import Annotated

import pytest
from hopscotch import Registry
from hopscotch.fixtures.dataklasses import Greeting
from hopscotch.operators import Context
from hopscotch.operators import context
from hopscotch.operators import Get
from hopscotch.operators import make_field_operator


def test_get_setup() -> None:
    """Setup a ``Get`` operator."""
    get = Get(Greeting)
    assert Greeting == get.lookup_key
    assert None is get.attr


def test_get_setup_attr() -> None:
    """Setup a ``Get`` operator with an attr."""
    get = Get(Greeting, attr="salutation")
    assert Greeting == get.lookup_key
    assert "salutation" == get.attr


def test_operators_get_simple_type() -> None:
    """Use Annotated to lookup one type while stating another."""

    @dataclass()
    class DummyGreeter:
        dummy_greeting: Annotated[Greeting, Get(Greeting)]

    registry = Registry()
    greeting = Greeting()
    registry.register(greeting)
    registry.register(DummyGreeter)
    # Now ask the registry to construct this
    dummy_greeter = registry.get(DummyGreeter)
    assert isinstance(dummy_greeter.dummy_greeting, Greeting)


def test_operators_get_attr() -> None:
    """Use Get with attr to pluck an attribute off the injected target."""

    @dataclass()
    class DummyGreeter:
        dummy_salutation: Annotated[str, Get(Greeting, "salutation")]

    registry = Registry()
    greeting = Greeting()
    registry.register(greeting)
    registry.register(DummyGreeter)
    # Now ask the registry to construct this
    dummy_greeter = registry.get(DummyGreeter)
    assert "Hello" == dummy_greeter.dummy_salutation


def test_operators_get_failed_string() -> None:
    """Operators can never look up a string in the registry."""
    registry = Registry()
    get = Get("Failure")
    expected = "Cannot use a string 'Failure' as container lookup value"
    with pytest.raises(ValueError) as exc:
        get(registry)
    assert exc.value.args[0] == expected


def test_operators_get_value_none() -> None:
    """Registry did not have the lookup key registered as a kind."""
    registry = Registry()
    get = Get(Greeting)
    with pytest.raises(LookupError) as exc:
        get(registry)
    expected = "No kind 'Greeting' in registry"
    assert exc.value.args[0] == expected


def test_operators_context_setup() -> None:
    """Ensure the context operator can be constructed."""
    context = Context(attr="title")
    assert context.attr == "title"


def test_operators_context() -> None:
    """Get the current registry context."""

    @dataclass
    class Context1:
        title: str = "Context 1"

    @dataclass
    class DummyHeading:
        context1: Annotated[Context1, Context()]

    context1 = Context1()
    registry = Registry(context=context1)
    registry.register(DummyHeading)
    result = registry.get(DummyHeading)
    assert result.context1 == context1


def test_operators_context_attr() -> None:
    """Pluck an attr off the registry context."""

    @dataclass
    class Context1:
        title: str = "Context 1"

    @dataclass
    class DummyHeading:
        context_title: str = context(attr="title")

    context1 = Context1()
    registry = Registry(context=context1)
    registry.register(DummyHeading)
    result = registry.get(DummyHeading)
    assert result.context_title == context1.title


def test_operators_operators_value_none() -> None:
    """Registry did not have the lookup key."""
    context = Context()
    with pytest.raises(ValueError) as exc:
        context(Registry())
    expected = "No context on registry"
    assert exc.value.args[0] == expected


def test_make_field_operator() -> None:
    """Turn an operator into a dataclass field."""
    get = make_field_operator(Get)
    get_field = get(Greeting, attr="salutation", metadata={"flag": 9})
    metadata = get_field.metadata
    assert metadata["flag"] == 9
    injected = metadata["injected"]
    operator: Get = injected["operator"]
    assert operator.lookup_key == Greeting
    assert operator.attr == "salutation"



"""Replace a built-in class."""
from dataclasses import dataclass

from hopscotch import injectable
from hopscotch import Registry


@dataclass
class Customer:
    """A basic customer."""

    name: str = "Mary"


@dataclass
class FrenchCustomer:
    """A different customer."""

    name: str = "Marie"


@injectable()
@dataclass
class Greeter:
    """A simple greeter."""

    greeting: str = "Hello!"


# Some site might want to change a built-in.
@injectable(kind=Greeter, context=FrenchCustomer)
@dataclass
class FrenchGreeter:
    """Provide a different ``Greeter`` in this site."""

    greeting: str = "Bonjour!"


customer = Customer()
french_customer = FrenchCustomer()
parent_registry = Registry(context=customer)
parent_registry.scan()
# Later
greeter1 = parent_registry.get(Greeter)
# greeter1.greeting == "Hello!"

# Much later
child_registry = Registry(parent=parent_registry, context=french_customer)
greeter2 = child_registry.get(Greeter)
# greeter2.greeting == "Bonjour!"



"""Example objects and kinds implemented as plain classes."""
from typing import Annotated
from typing import Optional


class Greeting:
    """A plain-old-class to give a greeting."""

    salutation: str = "Hello"


class GreetingNoDefault:
    """A plain-old-class to give a greeting without a default."""

    salutation: str = "Hello"

    def __init__(self, salutation: str):
        """Store salutation on the instance."""
        self.salutation = salutation


# Start Greeter
class Greeter:
    """A plain-old-class to engage a customer."""

    greeting: Greeting

    def __init__(self, greeting: Greeting):
        """Construct a greeter."""
        self.greeting = greeting


class GreeterKind:
    """A plain-old-class ``Kind`` to engage a customer."""

    greeting: Greeting

    def __init__(self, greeting: Greeting):
        """Construct a greeter that is a kind."""
        self.greeting = greeting


class GreeterOptional:
    """A plain-old-class to engage a customer with optional greeting."""

    greeting: Optional[Greeting]

    def __init__(self, greeting: Optional[Greeting]):
        """Construct a greeter with an optional greeting."""
        self.greeting = greeting


class GreeterAnnotated:
    """A plain-old-class to engage a customer with an annotation."""

    greeting: Annotated[Greeting, "YOLO"]

    def __init__(self, greeting: Greeting):
        """Construct a greeter that uses Annotated."""
        self.greeting = greeting


class GreeterChildren:
    """A plain-old-class that is passed a tree of VDOM nodes."""

    children: tuple[str]

    def __init__(self, children: tuple[str]):
        """Construct a greeter that has children."""
        self.children = children



"""Sample package with a __init__.py for discovery.

Used for testing caller_package and the branching that looks
if the package has a __init__.py in it.
"""
from typing import Any

from hopscotch.callers import caller_package


def make_call() -> Any:
    """Simulate something that is called and looks back for caller."""
    return caller_package()


def main_package() -> Any:
    """Simulate something that does the calling."""
    return make_call()



"""Make sure the test/example/docs examples work."""
from hopscotch import Registry
from hopscotch.fixtures import dataklasses
from hopscotch.fixtures import DummyOperator
from hopscotch.fixtures import functions
from hopscotch.fixtures import init_caller_package
from hopscotch.fixtures import named_tuples
from hopscotch.fixtures import plain_classes


def test_init_caller_package() -> None:
    """Test the fixture at tests.fixtures.init_caller_package."""
    result = init_caller_package.main_package()
    assert init_caller_package == result


def test_dataklass_fixtures() -> None:
    """Check the dataclass examples."""
    greeting = dataklasses.Greeting()
    assert greeting.salutation == "Hello"
    assert dataklasses.GreetingNoDefault(salutation="Hi").salutation == "Hi"
    assert not hasattr(dataklasses.GreetingInitFalse(), "salutation")
    assert dataklasses.GreetingOperator(greeting).greeter.salutation == "Hello"
    assert dataklasses.GreetingTuple(("Hi", "Hey")).salutation == ("Hi", "Hey")
    gs = dataklasses.Greeting()
    assert gs.salutation == "Hello"
    assert dataklasses.Greeter(greeting=greeting).greeting.salutation == "Hello"
    assert dataklasses.GreeterKind(greeting=gs).greeting.salutation == "Hello"
    gs2 = dataklasses.GreeterAnnotated(greeting=greeting).greeting.salutation
    assert gs2 == "Hello"
    children = ("a",)
    assert dataklasses.GreeterChildren(children=children).children == children
    assert dataklasses.GreeterOptional(greeting=None).greeting is None
    gi = dataklasses.AnotherGreeting()
    assert gi.salutation == "Another Hello"
    assert isinstance(gi, dataklasses.Greeting)


def test_functions_fixtures() -> None:
    """Check the function examples."""
    greeting = functions.Greeting()
    assert greeting == "Hello"
    assert functions.Greeter(greeting=greeting) == "Hello"
    assert functions.GreetingDefaultNoHint() == "Hello"
    assert functions.GreetingNoDefault(salutation="Hi") == "Hi"
    assert functions.Greeter(greeting=greeting) == "Hello"
    assert functions.GreeterAnnotated(customer_name="Hello") == "Hello"
    children = ("a",)
    assert functions.GreeterChildren(children=children) == children
    assert functions.GreeterOptional(greeting=None) is None


def test_named_tuples_fixtures() -> None:
    """Check the named tuple examples."""
    greeting = named_tuples.Greeting()
    assert greeting.salutation == "Hello"
    assert named_tuples.GreetingNoDefault(salutation="Hi").salutation == "Hi"
    assert named_tuples.Greeter(greeting=greeting).greeting.salutation == "Hello"
    greeter_annotated = named_tuples.GreeterAnnotated(greeting=greeting)
    assert greeter_annotated.greeting.salutation == "Hello"
    children = ("a",)
    assert named_tuples.GreeterChildren(children=children).children == children
    assert named_tuples.GreeterOptional(greeting=None).greeting is None


def test_plain_classes_fixtures() -> None:
    """Check the plain-old-classes examples."""
    greeting = plain_classes.Greeting()
    assert greeting.salutation == "Hello"
    assert plain_classes.GreetingNoDefault(salutation="Hi").salutation == "Hi"
    assert plain_classes.Greeter(greeting=greeting).greeting.salutation == "Hello"
    assert plain_classes.GreeterKind(greeting=greeting).greeting.salutation == "Hello"
    greeter_annotated = plain_classes.GreeterAnnotated(greeting=greeting)
    assert greeter_annotated.greeting.salutation == "Hello"
    children = ("a",)
    assert plain_classes.GreeterChildren(children=children).children == children
    assert plain_classes.GreeterOptional(greeting=None).greeting is None


def test_context() -> None:
    """Ensure the ``Customer`` for ``registry.context`` can be constructed."""
    customer = dataklasses.Customer(first_name="Mary")
    assert customer.first_name == "Mary"


def test_french_context() -> None:
    """Ensure construction of ``FrenchCustomer`` for ``registry.context``."""
    customer = dataklasses.FrenchCustomer(first_name="Marie")
    assert customer.first_name == "Marie"


def test_greeter_customer() -> None:
    """Ensure construction of ``GreeterCustomer``."""
    customer = dataklasses.Customer(first_name="Mary")
    greeter = dataklasses.GreeterCustomer(customer=customer)
    assert greeter.customer.first_name == "Mary"


def test_greeter_french_customer() -> None:
    """Ensure construction of ``GreeterFrenchCustomer``."""
    customer = dataklasses.FrenchCustomer(first_name="Marie")
    greeter = dataklasses.GreeterFrenchCustomer(customer=customer)
    assert greeter.customer.first_name == "Marie"


def test_dummy_get() -> None:
    """Check the fake operator."""
    dg = DummyOperator(arg="Hi")
    registry = Registry()
    assert dg(registry) == "Hi"
